# General purpose utilities

from numpy import *
import pylab as pl

import swan.gui
from swan import pycwt
swanrgb = swan.gui.swanrgb


def log2(x):
    """Base 2 logarithm"""
    return log(x)/log(2.)

def make_progress_str(i,L=100):
    """Make a progress string with a rotating stick"""
    return '\r%3d%s %s' % (100*i/L, '%', ('|','/','-','\\')[i%4])


def deriv(x,y):
    """Calculate derivative of a signal"""
    return  .5*((y[1:-1] - y[0:-2]) / (x[1:-1] - x[0:-2]) + \
            (y[2:]   - y[1:-1]) / (x[2:]   - x[1:-1]))

def get_extr_deriv(x,y,maxf=True,wth=4):
    """Derivative-based extrema detection"""
    der = deriv(x,y)
    a = der > 0
    if maxf:
        a = a[1:] < a[:-1]
    else:
        a = a[1:] > a[:-1]

    k = arange(len(a))[a]
    xe = ones(len(k),dtype='d')
    ye = ones(len(k),dtype='d')

    for j,i in enumerate(k):
        p = polyfit(x[i+1:i+3], der[i:i+2], 1)
        xe[j] = -p[1] / p[0] #???
        p = polyfit(x[i:i+3], y[i:i+3], 2)
        #ye[j] = polyval(p, x[i+1:i+2])[0]
        ye[j] = polyval(p, (xe[j],) )[0]
    return xe,ye

def detrend1(vect, n=1):
    j = arange(len(vect))
    return vect - polyval(polyfit(j, vect, n), j)

def deoutburst(vect, n = 4):
    d = std(vect)
    m = mean(vect)

    low = m - n*d
    high = m+ n*d

    res = vect.copy()
    for i in xrange(len(vect)):
        if vect[i] < low:
            res[i] = low
        elif vect[i] > high:
            res[i] = high
        else:
            res[i] = vect[i]
    return res

def default_freqs(Ns, f_s, num=100):
    """
    Return default frequencies vector
    -- Ns:  number of samples in data vector
    -- f_s: sampling frequency
    -- num: number of frequencies required
    """
    T = Ns/f_s
    return linspace(4/T, f_s/2, num=num)

def in_subplots(seq):
    figure();
    N = len(seq)
    for k,s in enumerate(seq):
        subplot(N,1,k+1);
        plot(s)

def ifnot(a, b):
    "if a is not None, return a, else return b"
    if a == None: return b
    else: return a


def DFoSD(vec, normL=None, th = 1e-6):
    "Remove mean and normalize to S.D."
    normL = ifnot(normL, len(vec))
    m, x  = np.mean, vec[:normL]
    sdx = np.std(x,0)
    out =  np.zeros(vec.shape, vec.dtype)
    if sdx.shape is ():
	if np.abs(sdx) > th:
		out = (vec-m(x))/sdx
    else:
	zi = np.where(np.abs(sdx) < th)[0]
	sdx[zi] = 1.0
	out = (vec-m(x))/sdx
	out[zi]=0
    return out

def confidence_contour(esurf, extent, ax, L=3.0):
    # Show 95% confidence level (against white noise, v=3 \sigma^2)
    ax.contour(esurf, [L], extent=extent,
               cmap=mpl.cm.gray)

def cone_infl(freqs, extent, wavelet, ax):
    try:
        ax.fill_betweenx(freqs,
                         extent[0],
                         extent[0]+wavelet.cone(freqs),
                         alpha=0.5, color='black')
        ax.fill_betweenx(freqs,
                         extent[1]+wavelet.cone(-freqs),
                         extent[1],
                         alpha=0.5, color='black')
    except:
        print("Can't use fill_betweenx function: update\
        maptlotlib?")


def wavelet_specgram(signal, f_s, freqs,  ax,
                     wavelet = pycwt.Morlet(),
                     padding = 'zpd',
                     cax = None,
                     vmin=None, vmax=None,
                     correct = None,
		     cwt_fn = pycwt.eds,
                     confidence_level = False):
    wcoefs = pycwt.cwt_f(signal, freqs, f_s, wavelet, padding)
    surf = cwt_fn(wcoefs, wavelet.f0)
    if vmax is None: vmax = percentile(surf, 99.0)
    if vmin is None: vmin = percentile(surf, 1.0)
    
    if correct == 'freq1':
        coefs = freqs*2.0/np.pi
        for i in xrange(surf.shape[1]):
            surf[:,i] *= coefs
    endtime = len(signal)/f_s
    extent=[0, endtime, freqs[0], freqs[-1]]
    im = ax.imshow(surf, extent = extent,
                   origin = 'low',
                   vmin = vmin, vmax = vmax,
                   cmap = swanrgb,
                   alpha = 0.95)
    if not cax:
        pl.colorbar(im, ax=ax)
    else:
        pl.colorbar(im, cax = cax)
    cone_infl(freqs, extent, wavelet, ax)
    if confidence_level:
        confidence_contour(eds, extent, ax, confidence_level)
    return surf

def setup_axes_for_spectrogram(figsize = (12,6)):
    "Set up axes for a plot with signal, spectrogram and a colorbar"
    fig = pl.figure(figsize = figsize)
    ax = [fig.add_axes((0.08, 0.4, 0.8, 0.5))]
    ax.append(fig.add_axes((0.08, 0.07, 0.8, 0.3), sharex=ax[0]))
    ax.append(fig.add_axes((0.9, 0.4, 0.02, 0.5), 
                           xticklabels=[], 
                           yticklabels=[]))
    return fig,ax



def plot_spectrogram_with_ts(signal, f_s, freqs=None,
                             figsize=(12,6),
                             lc = 'b', title_string = '',
                             **kwargs):
    "Create a figure of a signal, spectrogram and a colorbar"
    Ns = len(signal)*1.0
    freqs = ifnot(freqs, default_freqs(Ns, f_s,512))
    tvec = np.arange(0, (Ns+2)/f_s, 1./f_s)[:Ns]

    fig,axlist = setup_axes_for_spectrogram(figsize)

    axlist[1].plot(tvec, signal,'-',color=lc)

    kwargs['cax'] = axlist[2]
    surf = wavelet_specgram(signal, f_s, freqs,  axlist[0], **kwargs)
    axlist[0].set_title(title_string)
    axlist[0].axis((tvec[0],tvec[-1], freqs[0],freqs[-1]))
    #axlist[1].xlim((tvec[0],tvec[-1]))
    return fig, surf




# Continuous wavelet transfrom via Fourier transform
# Collection of routines for wavelet transform via FFT algorithm


#-- Some naming and other conventions --
# use f instead of omega wherever rational/possible
# *_ft means Fourier transform

#-- Some references --
# [1] Mallat, S.  A wavelet tour of signal processing
# [2] Addison, Paul S. The illustrated wavelet transform handbook
# [3] Torrence and Compo. A practical guide to wavelet
#     analysis. Bulletin of American Meteorological Society. 1998, 79(1):61-78

import numpy as np
from numpy.fft import fft, ifft, fftfreq

_maxshape_ = 5e6

def memsafe_arr(shape, dtype):
    import tempfile as tmpf
    N = shape[0] * shape[1]
    if N < _maxshape_:
        return np.zeros(shape, dtype=dtype)
    else:
        print "Using memory-mapped arrays..."
        _tmpfile = tmpf.TemporaryFile('w+')
        return np.memmap(_tmpfile, dtype=dtype, shape=shape)
        
    

#try:
#    from scipy.special import gamma
#except:

pi = np.pi

try:
    from scipy.special import gamma
    class DOG:
        """Derivative of Gaussian, general form"""
        # Incomplete, as the general form of the mother wavelet
        # would require symbolic differentiation.
        # Should be enough for the CWT computation, though

        def __init__(self, m = 1.):
            self.order = m
            self.fc = (m+.5)**.5 / (2*pi)
            self.f0 = self.fc
    
        def psi_ft(self, f):
            c = 1j**self.order / np.sqrt(gamma(self.order + .5)) #normalization
            w = 2*pi*f
            return c * w**self.order * np.exp(-.5*w**2)
except:
    class DOG:
        def __init__(self,*args):
            print "Couldn't create DOG wavelet because scipy is unavailable"
            return None
        

class Mexican_hat:
    def __init__(self, sigma = 1.0):
        self.sigma = sigma
        self.fc = .5 * 2.5**.5 / pi
        self.f0 = self.fc
    def psi_ft(self, f):
        """Fourier transform of the Mexican hat wavelet"""
        c = np.sqrt(8./3.) * pi**.25 * self.sigma**2.5 
        wsq = (2. * pi * f)**2.
        return -c * wsq * np.exp(-.5 * wsq * self.sigma**2.)
    def psi(self, tau):
        """Mexian hat wavelet as described in [1]"""
        xsq = (tau / self.sigma)**2.
        c = 2 * pi**-.25 / np.sqrt(3 * self.sigma) # normalization constant from [1]
        out =  c * (1 - xsq) * np.exp(-.5 * xsq)
        out *= (tau < 6) * (tau > -6) # make it 'compact support'
        return out
    def cone(self, f):
        "e-folding time [Torrence&Compo]. For frequencies"
        return self.f0*2.0**0.5 / f
    def cone_s(self, s):
        "e-folding time [Torrence&Compo]. For scales"
        return 2.**0.5*s
    def set_f0(self, f0):
        pass

def heavyside(x):
    return 1.0*(x>0.0)

class Morlet:
    def __init__(self, f0 = 1.5):
        self.set_f0(f0)
    def psi(self, t):
        pi**0.25 * exp(-t**2 / 2.) * exp(2j*pi*self.f0*t) #[3]
    def psi_ft(self, f):
        """Fourier transform of the approximate Morlet wavelet
            f0 should be more than 0.8 for this function to be
            correct."""
        # [3]
        coef = (pi**-.25) * heavyside(f)
        return  coef * np.exp(-.5 * (2. * pi * (f - self.f0))**2.)
    def sigma_t_s(self,s):
        """
        Heisenberg box width for scales [Addison]
        s/sqrt(2)

        """
        return s/np.sqrt(2)

    def sigma_f_s(self, s):
        """
        Heisenberg box height for scales [Addison]
        s*sqrt(2)/4
        """
        return 1.0/(s*pi*np.sqrt(8.0))

    def sigma_f_f(self,f):
        """
        Heisenberg box height in frequencies
        s*sqrt(2)/4
        """
        return self.sigma_f_s(self.f0/f)

    def cone(self, f):
        "e-folding time [Addison]. For frequencies"
        return self.f0/(f*2.0**0.5)
    def cone_s(self, s):
        "e-folding time [Addison]. For scales"
        return self.sigma_t_s(s)
    def set_f0(self, f0):
        self.f0 = f0
        #self.fc = f0


def next_pow2(x):
    return 2.**np.ceil(np.log2(x))

def pad_func(ppd):
    func = None
    if ppd == 'zpd':
        func = lambda x,a:  x*0.0
    elif ppd == 'cpd':
        func = lambda x,a:  x*a
    return func
    
def evenp(n):
    return not n%2

def cwt_a(signal, scales, sampling_scale = 1.0,
          wavelet=Mexican_hat(),
          ppd = 'zpd',
          verbose = False,):
    """ Continuous wavelet transform via fft. Scales version."""
    import sys
    siglen = len(signal)

    if hasattr(wavelet, 'cone_s'):
        needpad = wavelet.cone_s(np.max(scales))
    else:
        needpad = 1.2*np.max(scales)

    if siglen < 5e5: # only do expansion up to next ^2 for short signals
        ftlen = int(next_pow2(siglen + 2*needpad))
    else:
        if not evenp(siglen) :
            siglen -= 1
            signal = signal[:-1]
        ftlen = int(siglen + 2*needpad)

    padlen1 = int((ftlen - siglen)/2)
    padlen2 = evenp(siglen) and padlen1 or padlen1+1
    
    padded_signal = np.ones(ftlen)
    padfunc = pad_func(ppd)
    padded_signal[:padlen1] = padfunc(padded_signal[:padlen1],signal[0])
    padded_signal[-padlen2:] = padfunc(padded_signal[-padlen2:],signal[-1])
    padded_signal[padlen1:-padlen2] = signal

    signal_ft = fft(padded_signal)     # FFT of the signal
    del padded_signal
    ftfreqs = fftfreq(ftlen, sampling_scale)  # FFT frequencies

    psi_ft = wavelet.psi_ft
    coef = np.sqrt(2*pi/sampling_scale)

    ### version with map is slower :(
    #def _ls(s):
    #    return ifft(signal_ft*coef*(s**0.5)*np.conjugate(psi_ft(s*ftfreqs)))[padlen:-padlen]
    #xx = map(_ls, scales)
    #W = np.array(xx)

    ### matrix version
    scales = scales.reshape(-1,1)
    #s_x_ftf = np.dot(scales, ftfreqs.reshape(1,-1))
    #psi_ft_bar = np.conjugate(psi_ft(s_x_ftf))
    #psi_ft_bar *= coef*np.sqrt(scales)
    #W = ifft( psi_ft_bar * signal_ft.reshape(1,-1))[:,padlen1:-padlen2]
    #return W

    ## create the result matrix beforehand
    #W = np.zeros((len(scales), siglen), 'complex')
    W = memsafe_arr((len(scales), siglen), 'complex')
    ## Now fill in the matrix
    for n,s in enumerate(scales):
        if verbose :
            sys.stderr.write('\r Processing scale %04d of %04d'%(n+1,len(scales)))
        psi_ft_bar = np.conjugate(psi_ft(s * ftfreqs))
        psi_ft_bar *= coef*np.sqrt(s) # Normalization from [3]
        W[n,:] = ifft(signal_ft * psi_ft_bar)[padlen1:-padlen2]
    return W


def cwt_f(signal, freqs, Fs=1.5, wavelet = Morlet(), ppd = 'zpd', verbose=False):
    """Continuous wavelet transform -- frequencies version"""
    scales = wavelet.f0/freqs
    dt = 1./Fs
    return cwt_a(signal, scales, dt, wavelet, ppd, verbose=verbose)


def eds(x, f0=1.5):
    "Energy density surface [2,3]"
    ## Update, simulations with MK (as in [3]) suggest that I
    ## shouldn't divide by f0 to obtain correct normalisation,
    ## e.g. 1 s.d. is mean for white noise signal
    #return abs(x)**2/f0
    return abs(x)**2

def real(x, *args):
    return x.real

def cwt_phase(x, *args):
    return np.arctan2(x.imag, x.real)


def xwt_f(sig1, sig2, freqs, Fs=1.0, wavelet=Morlet()):
    "Cross-wavelet coeficients for 2 signals"
    cwtf = lambda x: cwt_f(x, freqs, Fs, wavelet)
    return xwt(cwtf(sig1),cwtf(sig2))

def absxwt_f(sig1, sig2, freqs, Fs=1.0, wavelet=Morlet()):
    "Cross-wavelet power for 2 signals"
    return abs(xwt_f(sig1,sig2, freqs, Fs, wavelet))/wavelet.f0**0.5

def absxwt_a(sig1, sig2, scales, dt=1.0, wavelet=Morlet()):
    "Cross-wavelet power for 2 signals"
    Fs = 1./dt
    freqs = wavelet.f0/scales
    return abs(xwt_f(sig1,sig2, freqs, Fs, wavelet))/wavelet.f0**0.5


def xwt(wcoefs1,wcoefs2):
    "Cross wavelet transform for 2 sets of coefficients"
    return wcoefs1*wcoefs2.conjugate()

def absxwt(wcoefs1,wcoefs2, f0=1.5):
    "Cross-wavelet power for 2 sets of coefficients"
    ## Why do I divide by f_0^2 here?
    return  abs(xwt(wcoefs1,wcoefs2))/f0**0.5

def wtc_a(sig1, sig2, scales, dt=1.0, wavelet=Morlet()):
    cwta = lambda x: cwt_a(x, scales, dt, wavelet)
    return coherence_a(cwta(sig1), cwta(sig2), scales, dt, wavelet)


def wtc_f(sig1, sig2, freqs, Fs=1.0, wavelet=Morlet()):
    cwtf = lambda x: cwt_f(x, freqs, Fs, wavelet)
    return coherence_f(cwtf(sig1), cwtf(sig2), freqs, Fs, wavelet.f0)

def coherence_a(x,y,scales,dt,f0=1.5):
    # almost not useful 
    #scor =np.ones(x.shape[1])[:,np.newaxis] * scales
    #scor = np.transpose(scor)
    sx = wsmooth_a(abs(x)**2,scales,dt)
    sy= wsmooth_a(abs(y)**2,scales,dt)
    sxy = wsmooth_a((x*y.conjugate()), scales, dt)
    return abs(sxy)**2/(sx.real*sy.real)

def coherence_f(x,y,freqs, Fs=1.0, f0=1.5):
    return coherence_a(x,y,f0/freqs, 1.0/Fs, f0)

def cphase(x,y):
    #hard to interprete :(
    d = xwt(x,y)
    return np.arctan2(d.imag, d.real)



def wsmooth_a(coefs, scales, dt = 1.0, wavelet=Morlet()):
    "Smoothing of wavelet coefs. Scales version"
    #G = lambda omega,s: np.exp(-0.5 * s**2.0 *omega**2.0) # Literature
    G = lambda omega,s: np.exp(-0.25 * s**2.0 *omega**2.0) 

    W = np.zeros(coefs.shape, 'complex')
    fftom = 2*pi*fftfreq(coefs.shape[1], dt) 

    for n,s in enumerate(scales):
        W[n,:] = ifft(fft(coefs[n,:]) * G(fftom, s/dt))

    # TODO: scale smoothing
    return W


import time 
def speed_test(sig_len, nscales = 512, N = 100, ppd='zpd',wavelet=Morlet()):
    tick = time.clock()
    s = np.random.randn(sig_len)
    scales = np.linspace(0.5, sig_len/4.0, nscales)
    for j in xrange(N):
        eds = cwt_a(s, scales, sampling_scale = 1.0, ppd = ppd,
                    wavelet=wavelet)
    tock =  time.clock() - tick
    print tock
    return tock



# Discrete wavelet transform

import numpy as np

try:
    from scipy.ndimage import convolve1d
    _scipy_loaded = True
except :
    _scipy_loaded = False

def upsample(v):
    out = np.zeros(len(v)*2-1)
    out[::2] = v
    return out

def next_pow2(x):
    return 2.**np.ceil(np.log2(x))

def pad_func(ppd):
    func = None
    if ppd == 'zpd':
        func = lambda x,a:  x*0.0
    elif ppd == 'cpd':
        func = lambda x,a:  x*a
    return func
    
def evenp(n):
    return not n%2

def mirrorpd(k, L):
    if 0 <= k < L : return k
    else: return -(k)%L


# Redundant, modwt, a' trous
def dec_atrous(sig, lev, phi=np.array([1./16, 1./4, 3./8, 1./4, 1./16])):
    L = len(sig) 
    padlen = len(phi)
    assert L > padlen
    if _scipy_loaded:
        apprx = convolve1d(sig, phi, mode='mirror')
    else:
        indices = map(lambda i: mirrorpd(i, L),
                      range(-padlen, 0) + range(0,L) + range(L, L+padlen))
        padded_sig = sig[indices]
        apprx = np.convolve(padded_sig, phi, mode='same')[padlen:padlen+L]
    w = (sig - apprx) # wavelet coefs
    if lev <= 0: return sig
    elif lev == 1 or L < len(upsample(phi)): return [w, apprx]
    else: return [w] + dec_atrous(apprx, lev-1, upsample(phi))


# version with direct convolution (slow)
def dec_atrous1(v, lev, phi=np.array([1./16, 1./4, 3./8, 1./4, 1./16])):
    coefs = []
    cprev = v.copy()
    cnext = np.zeros(v.shape)
    L,lphi = len(v), len(phi)
    phirange = np.arange(lphi) - int(lphi/2)
    Ll = xrange(L)
    for j in xrange(lev):
        phiind = (2**j)*phirange
	cvals = [np.sum(phi*cprev[(l+phiind)%L]) for l in  Ll]
        coefs.append(cprev - cvals)
        cprev = np.array(cvals)
    return coefs + [cprev]
